#include "kinectpointsource.hpp"
#include <cstring>
#include <iostream>

#define unlikely(x) __builtin_expect((x),0)

KinectPointSource::KinectPointSource(const std::string& yml)
    : yml(yml) {
    
    initialize_z_lookup();
    initialize_matrices();
}

void KinectPointSource::initialize_z_lookup() {
    for(int i = 0; i < 2048; i++)
        z_lookup.push_back(1.0f / (i * -0.0030711016f + 3.3309495161f));
}

void KinectPointSource::initialize_matrices() {
    cv::Mat rgb_intrinsics, depth_intrinsics;
    cv::Mat rgb_distortion, depth_distortion;
    
    cv::FileStorage storage(yml, cv::FileStorage::READ);
    storage["rgb_intrinsics"] >> rgb_intrinsics;
    storage["rgb_distortion"] >> rgb_distortion;
    storage["depth_intrinsics"] >> depth_intrinsics;
    storage["depth_distortion"] >> depth_distortion;
    
    cv::Mat tmp_r, tmp_t;
    storage["R"] >> tmp_r;
    storage["T"] >> tmp_t;
    for(int i = 0; i < 3; i++) {
        R.row(i) << tmp_r.at<double>(i, 0),
                    tmp_r.at<double>(i, 1),
                    tmp_r.at<double>(i, 2);
        
        T(i) = tmp_t.at<double>(i, 0);
    }
    
    // Erzeuge die UndistortMap für die Farbkamera
    {
        cv::Mat map1, map2;
        cv::initUndistortRectifyMap(rgb_intrinsics, rgb_distortion,
            cv::Mat(), rgb_intrinsics, cv::Size(640, 480),
            CV_32FC1, map1, map2);

        cv::convertMaps(map1, map2, rgb_map_x, rgb_map_y, CV_16SC2);
    }
    
    // Erzeuge die UndistortMap für die Tiefenkamera
    {
        cv::Mat map1, map2;
        cv::initUndistortRectifyMap(depth_intrinsics, depth_distortion,
            cv::Mat(), depth_intrinsics, cv::Size(640, 480),
            CV_32FC1, map1, map2);
    
        cv::convertMaps(map1, map2, depth_map_x, depth_map_y, CV_16SC2);
    }
    
    fx_d = depth_intrinsics.at<double>(0, 0);
    fy_d = depth_intrinsics.at<double>(1, 1);
    cx_d = depth_intrinsics.at<double>(0, 2);
    cy_d = depth_intrinsics.at<double>(1, 2);
    
    fx_rgb = rgb_intrinsics.at<double>(0, 0);
    fy_rgb = rgb_intrinsics.at<double>(1, 1);
    cx_rgb = rgb_intrinsics.at<double>(0, 2);
    cy_rgb = rgb_intrinsics.at<double>(1, 2);
}

void KinectPointSource::remap_depth(const cv::Mat& src, cv::Mat& dst) const {
    cv::remap(src, dst, depth_map_x, depth_map_y, CV_INTER_NN);
}

void KinectPointSource::remap_rgb(const cv::Mat& src, cv::Mat& dst) const {
    cv::remap(src, dst, rgb_map_x, rgb_map_y, CV_INTER_LINEAR);
}

void KinectPointSource::reproject(const cv::Mat& rgb, const cv::Mat& depth,
        std::vector<cv::Point3f> &points,
        std::vector<cv::Vec3b> &colors) const {
    
    const int width = rgb.cols;
    const int height = rgb.rows;
    const uint16_t *p_depth = reinterpret_cast<uint16_t*>(depth.data);
    
    points = std::vector<cv::Point3f>(width*height);
    colors = std::vector<cv::Vec3b>(width*height);
    
    #pragma omp parallel for
    for(int y = 0; y < height; ++y) {
        const int y_offset = y*width;
        for(int x = 0; x < width; ++x) {
            cv::Point3f point;
            point.z = z_lookup[p_depth[y_offset+x] & 0x7ff];
            point.x = (x - cx_d) * point.z / fx_d;
            point.y = (y - cy_d) * point.z / fy_d;
            
            Eigen::Vector3f vec(point.x, point.y, point.z);
            vec = R * vec - T;
            
            int cx = static_cast<int>(vec(0) * fx_rgb / vec(2) + cx_rgb + 0.5f);
            int cy = static_cast<int>(vec(1) * fy_rgb / vec(2) + cy_rgb + 0.5f);
            if(unlikely(cx < 0 || cx >= width || cy < 0 || cy >= height))
                continue;
            
            points[y_offset+x] = point;
            colors[y_offset+x] = rgb.at<cv::Vec3b>(cy, cx);
        }
    }
}

